@@ -3,6 +3,7 @@ from .adapters import adapters
 from .models import models
 from .parsing import parsing
 from .rs2wapy import RS2WebAdmin
+from .epicgamesstore import epicgamesstore
 
 __version__ = get_versions()["version"]
 del get_versions
@@ -12,4 +13,5 @@ __all__ = [
     "models",
     "parsing",
     "RS2WebAdmin",
+    "epicgamesstore",
 ]
