@@ -2,10 +2,12 @@ from .adapters import BanWrapper
 from .adapters import PlayerWrapper
 from .adapters import SquadWrapper
 from .adapters import WebAdminAdapter
+from .adapters import logger
 
 __all__ = [
     "BanWrapper",
     "WebAdminAdapter",
     "SquadWrapper",
     "PlayerWrapper",
+    "logger",
 ]
