@@ -14,7 +14,11 @@ from dataclasses import dataclass
 from http import HTTPStatus
 from io import BytesIO
 from pathlib import Path
+from typing import Any
+from typing import Callable
+from typing import Dict
 from typing import List
+from typing import Optional
 from typing import Sequence
 from typing import Tuple
 from typing import Type
@@ -58,15 +62,16 @@ BAN_EXP_NUMBER_MAX = 365
 BAN_EXP_NUMBERS = range(BAN_EXP_NUMBER_MAX)
 BAN_EXP_PATTERN = re.compile(r"([0-9]*)\s?(\L<units>)",
                              units=[b.lower() for b in BAN_EXP_UNITS])
+WEB_ADMIN_DATE_FMT = "%Y%m%d %H%M"
 BAN_ID_TYPE_STEAM_ID_64 = 1
 
 MAP_PREFIX_TO_GAME_TYPE = {
     "VNTE": "ROGame.ROGameInfoTerritories",
     "VNSU": "ROGame.ROGameInfoSupremacy",
     "VNSK": "ROGame.ROGameInfoSkirmish",
-    "WWTE": "WWGame.WWGameInfoTerritories",
-    "WWSU": "WWGame.WWGameInfoSupremacy",
-    "WWSK": "WWGame.WWGameInfoSkirmish",
+    "WWTE": "WinterWar.WWGameInfoTerritories",
+    "WWSU": "WinterWar.WWGameInfoSupremacy",
+    "WWSK": "WinterWar.WWGameInfoSkirmish",
     "GMTE": "GreenMenMod.GMGameInfoTerritories",
     "GMSU": "GreenMenMod.GMGameInfoSupremacy",
     "GMSK": "GreenMenMod.GMGameInfoSkirmish",
@@ -85,9 +90,10 @@ WEB_ADMIN_MAP_LIST_PATH = WEB_ADMIN_SETTINGS_PATH / Path("maplist/")
 WEB_ADMIN_BANS_PATH = WEB_ADMIN_POLICY_PATH / Path("bans/")
 WEB_ADMIN_SQUADS_PATH = WEB_ADMIN_CURRENT_GAME_PATH / Path("squads/")
 WEB_ADMIN_TRACKING_PATH = WEB_ADMIN_POLICY_PATH / Path("tracking/")
+WEB_ADMIN_MEMBERS_URL = WEB_ADMIN_POLICY_PATH / Path("members/")
 
 
-def retry(on_exc: Union[Exception, Tuple],
+def retry(on_exc: Union[Type[Exception], Tuple[Exception, ...]],
           tries: int = 10, delay: int = 3, backoff: int = 2,
           cap: int = 30):
     """Retry calling the decorated function
@@ -128,16 +134,16 @@ def _set_postfields(curl_obj: pycurl.Curl, postfields: str):
     postfieldsize = len(postfields)
     logger.debug("postfieldsize: {pf_size}", pf_size=postfieldsize)
     logger.debug("postfields: {pf}", pf=postfields)
-    curl_obj.setopt(curl_obj.POSTFIELDS, postfields)
-    curl_obj.setopt(curl_obj.POSTFIELDSIZE_LARGE, postfieldsize)
+    curl_obj.setopt(pycurl.POSTFIELDS, postfields)
+    curl_obj.setopt(pycurl.POSTFIELDSIZE_LARGE, postfieldsize)
 
 
 def _policies_to_delete_argstr(policies: List[str], to_delete: str) -> str:
     del_index = [idx for idx, s in enumerate(policies) if to_delete in s][0]
-    policies = [p.split(":") for p in policies]
+    policies_split = [p.split(":") for p in policies]
     policies = [(f"ipmask={p[0].strip()}&policy={p[1].strip()}"
                  f"{f'&delete={del_index}' if p[0] == to_delete else ''}")
-                for p in policies]
+                for p in policies_split]
     return "&".join(policies)
 
 
@@ -181,13 +187,14 @@ class WebAdminAdapter:
     }
 
     def __init__(self, username: str, password: str, webadmin_url: str):
-        self._headers = {}
+        self._headers: Dict[str, str] = {}
+        self._chat_message_deque: deque = deque(maxlen=512)
+        self._auth_data: Optional[AuthData] = None
+
         self._username = username
         self._webadmin_url = webadmin_url
-        self._password_hash = ""
+        self._password_hash = b""
         self._hash_alg = ""
-        self._auth_data = None
-        self._chat_message_deque = deque(maxlen=512)
 
         scheme, netloc, path, params, query, fragment = urlparse(self._webadmin_url)
         logger.debug("webadmin_url={url}, scheme={scheme}, netloc={netloc}, "
@@ -259,6 +266,10 @@ class WebAdminAdapter:
             (scheme, netloc, WEB_ADMIN_TRACKING_PATH.as_posix(),
              params, query, fragment)
         )
+        self._members_url = urlunparse(
+            (scheme, netloc, WEB_ADMIN_MEMBERS_URL.as_posix(),
+             params, query, fragment)
+        )
 
         # Perform a dummy request to populate self._headers.
         self._perform(self._webadmin_url, skip_auth=True)
@@ -280,7 +291,7 @@ class WebAdminAdapter:
             pass
 
     @property
-    def auth_data(self) -> AuthData:
+    def auth_data(self) -> Optional[AuthData]:
         return self._auth_data
 
     @auth_data.setter
@@ -288,14 +299,14 @@ class WebAdminAdapter:
         self._auth_data = auth_data
 
     @property
-    def _pw_hash(self) -> str:
-        return base64.decodebytes(self._password_hash).decode("utf-8")
+    def _pw_hash(self) -> bytes:
+        return base64.decodebytes(self._password_hash)
 
     @_pw_hash.setter
-    def _pw_hash(self, pw_hash: str):
+    def _pw_hash(self, pw_hash: bytes):
         # Avoid storing plaintext password in memory in case
         # WebAdmin did not set 'hashAlg'.
-        self._password_hash = base64.encodebytes(pw_hash.encode("utf-8"))
+        self._password_hash = base64.encodebytes(pw_hash)
 
     def get_current_game(self) -> models.CurrentGame:
         headers = self._make_auth_headers()
@@ -340,19 +351,20 @@ class WebAdminAdapter:
         self._chat_message_deque.extend(chat_msgs)
 
     def get_access_policy(self) -> List[str]:
-        sessionid = self._auth_data.sessionid
-        authcred = self._auth_data.authcred
-        authtimeout = self._auth_data.authtimeout
-
-        headers = self.BASE_HEADERS.copy()
-        headers["Cookie"] = f"{sessionid}; {authcred}; {authtimeout}"
-        headers["Cache-Control"] = "no-cache"
-
-        # Prevent caching with randomized parameter.
-        url = f"{self._access_policy_url}?$(date +%s)"
-
-        resp = self._perform(url, headers=headers)
-        return self._rparser.parse_access_policy(resp)
+        # sessionid = self._auth_data.sessionid
+        # authcred = self._auth_data.authcred
+        # authtimeout = self._auth_data.authtimeout
+        #
+        # headers = self.BASE_HEADERS.copy()
+        # headers["Cookie"] = f"{sessionid}; {authcred}; {authtimeout}"
+        # headers["Cache-Control"] = "no-cache"
+        #
+        # # Prevent caching with randomized parameter.
+        # url = f"{self._access_policy_url}?$(date +%s)"
+        #
+        # resp = self._perform(url, headers=headers)
+        # return self._rparser.parse_access_policy(resp)
+        raise NotImplementedError
 
     # TODO: Refactor.
     def add_access_policy(self, ip_mask: str, policy: str) -> bool:
@@ -539,10 +551,30 @@ class WebAdminAdapter:
 
     def kick_player(self, player: Union[models.Player, PlayerWrapper],
                     reason: str, notify_players: bool = False):
-        raise NotImplementedError
+        headers = self._make_auth_headers()
+        players_resp = self._perform(self._players_url, headers=headers)
+        player_id, player_key = self._rparser.parse_player_id_key(
+            players_resp, player)
+
+        postfields = {
+            "action": "kick",
+            "playerid": player_id,
+            "playerkey": player_key,
+            "__Submitter": "",
+            "__UniqueId": "",
+            "__PlayerName": "",
+            "__Reason": reason,
+            "__NotifyPlayers": int(notify_players),
+            "__IdType": 0,
+            "__ExpNumber": "",
+            "__ExpUnit": "Never",
+        }
+        self._perform(self._players_url, headers=headers,
+                      postfields=postfields)
 
     def ban_player(self, player: Union[models.Player, PlayerWrapper],
-                   reason: str, duration: str, notify_players: bool = False):
+                   reason: str, duration: Optional[str] = None,
+                   notify_players: bool = False):
         steam_id = player.steam_id.as_64
 
         name = player.persona_name
@@ -553,10 +585,9 @@ class WebAdminAdapter:
         exp_unit = BAN_EXP_UNIT_PERM
 
         if duration:
-            # noinspection PyBroadException
             try:
                 exp_number, exp_unit = self._parse_ban_duration(duration)
-            except Exception:
+            except ValueError:
                 logger.warning("invalid ban duration")
 
         postfields = {
@@ -634,16 +665,43 @@ class WebAdminAdapter:
         return self._rparser.parse_squads(resp, adapter=self)
 
     def get_tracked_players(self) -> List[TrackingWrapper]:
-        headers = self._make_auth_headers()
-        resp = self._perform(self._tracking_url, headers=headers)
-        tracked_players = []
-        for _ in range(1):
-            tracked_players.extend(
-                self._rparser.parse_tracking(resp, adapter=self))
+        return self._get_multi_page_content(
+            url=self._tracking_url,
+            parse_func=self._rparser.parse_tracking,
+        )
 
-        # Remove duplicates in case tracking tab was
+    def get_members(self) -> List[MemberWrapper]:
+        return self._get_multi_page_content(
+            url=self._members_url,
+            parse_func=self._rparser.parse_members,
+        )
+
+    def _get_multi_page_content(
+            self,
+            url: str,
+            parse_func: Callable[[bytes, WebAdminAdapter], List[Any]],
+    ) -> List[Any]:
+        headers = self._make_auth_headers()
+
+        resp = self._perform(url, headers=headers)
+
+        content = parse_func(resp, self)
+        has_next_page = self._rparser.parse_has_next_page(resp)
+
+        while has_next_page:
+            fvri = self._rparser.parse_fvri(resp)
+            logger.debug("first visible row index: {fvri}", fvri=fvri)
+            postfields = {
+                "action": "nextpage",
+                "__FirstVisibleRowIndex": fvri,
+            }
+            resp = self._perform(url, headers=headers, postfields=postfields)
+            content.extend(parse_func(resp, self))
+            has_next_page = self._rparser.parse_has_next_page(resp)
+
+        # Remove duplicates in case content tab was
         # mutated during parsing.
-        return list(set(tracked_players))
+        return list(set(content))
 
     def _enqueue_chat_messages(self):
         while True and not self._stop_event.is_set():
@@ -671,15 +729,15 @@ class WebAdminAdapter:
             curl_obj = pycurl.Curl()
 
         if postfields:
-            postfields = urlencode(postfields)
-            _set_postfields(curl_obj, postfields)
+            postfields_dict = urlencode(postfields)
+            _set_postfields(curl_obj, postfields_dict)
 
         logger.debug("url={url}, headers={headers}", url=url, headers=headers)
         if not headers:
             headers = self.BASE_HEADERS.copy()
-        headers = self._headers_to_list(headers)
+        headers_list = self._headers_to_list(headers)
 
-        logger.debug("prepared headers={h}", h=headers)
+        logger.debug("prepared headers={h}", h=headers_list)
 
         buffer = BytesIO()
 
@@ -687,7 +745,7 @@ class WebAdminAdapter:
         curl_obj.setopt(pycurl.HEADERFUNCTION, self._header_function)
         curl_obj.setopt(pycurl.BUFFERSIZE, 102400)
         curl_obj.setopt(pycurl.URL, url)
-        curl_obj.setopt(pycurl.HTTPHEADER, headers)
+        curl_obj.setopt(pycurl.HTTPHEADER, headers_list)
         curl_obj.setopt(pycurl.USERAGENT, CURL_USERAGENT)
         curl_obj.setopt(pycurl.MAXREDIRS, 50)
         curl_obj.setopt(pycurl.ACCEPT_ENCODING, "")
@@ -814,10 +872,11 @@ class WebAdminAdapter:
         pw_hash = ""
         pw = ""
 
+        pw_hash_str = self._pw_hash.decode("utf-8")
         if self._hash_alg:
-            pw_hash = f"${self._hash_alg}${self._pw_hash}"
+            pw_hash = f"${self._hash_alg}${pw_hash_str}"
         else:
-            pw = self._pw_hash
+            pw = pw_hash_str
 
         postfields = {
             "token": token,
@@ -888,11 +947,12 @@ class WebAdminAdapter:
             logger.exception(ae)
 
     def _make_auth_headers(self) -> dict:
-        sessionid = self._auth_data.sessionid
-        authcred = self._auth_data.authcred
-        authtimeout = self._auth_data.authtimeout
         headers = self.BASE_HEADERS.copy()
-        headers["Cookie"] = f"{sessionid}; {authcred}; {authtimeout}"
+        if self._auth_data is not None:
+            sessionid = self._auth_data.sessionid
+            authcred = self._auth_data.authcred
+            authtimeout = self._auth_data.authtimeout
+            headers["Cookie"] = f"{sessionid}; {authcred}; {authtimeout}"
         return headers
 
     def _make_chat_headers(self) -> dict:
@@ -906,10 +966,10 @@ class WebAdminAdapter:
         self._hash_alg = self._rparser.parse_hash_alg(resp)
         logger.debug("using hash algorithm: '{a}'", a=self._hash_alg)
         if self._hash_alg:
-            self._pw_hash = self._ue3_pw_hash_digest(username, password)
+            self._pw_hash = self._pw_hash_digest(username, password)
         else:
             # Hash algorithm was not set.
-            self._pw_hash = password
+            self._pw_hash = password.encode("utf-8")
 
     def _read_encoding(self, index: int = -1) -> str:
         encoding = None
@@ -927,14 +987,21 @@ class WebAdminAdapter:
             logger.debug("assuming encoding is {enc}", enc=encoding)
         return encoding
 
-    def _ue3_pw_hash_digest(self, username: str, password: str) -> str:
+    def _pw_hash_digest(self, username: str, password: str) -> bytes:
         """
-        Calculate the hex digest used by Unreal Engine 3 WebAdmin,
-        which is transmitted over the wire.
+        Calculate the hex digest used by Rising Storm 2: Vietnam
+        WebAdmin, which is transmitted over the wire.
         """
-        hash_alg = getattr(hashlib, self._hash_alg)
-        return hash_alg(bytearray(password, "utf-8")
-                        + bytearray(username, "utf-8")).hexdigest()
+        try:
+            hash_alg = getattr(hashlib, self._hash_alg)
+            return hash_alg(bytearray(password, "utf-8")
+                            + bytearray(username, "utf-8")
+                            ).hexdigest().encode("utf-8")
+        except AttributeError as ae:
+            logger.debug(ae, exc_info=True)
+            logger.error("hash algorithm '{ha}' is not supported",
+                         ha=self._hash_alg)
+            raise
 
     @staticmethod
     def _headers_to_list(headers: dict) -> List[str]:
@@ -944,7 +1011,7 @@ class WebAdminAdapter:
         return [f"{key}: {value}" for key, value in headers.items()]
 
     @staticmethod
-    def _parse_ban_duration(duration: str) -> Tuple[int, str]:
+    def _parse_ban_duration(duration: str) -> Tuple[str, str]:
         try:
             duration = duration.lower().strip()
             m = re.match(BAN_EXP_PATTERN, duration)
@@ -955,12 +1022,12 @@ class WebAdminAdapter:
             duration_unit = m.group(2)
             logger.debug("duration_unit: {du}", du=duration_unit)
             duration_unit = duration_unit[0].upper() + duration_unit[1:]
-            return ban_duration, duration_unit
+            return str(ban_duration), duration_unit
         except Exception as e:
             logger.debug("error parsing ban duration string: {e}",
                          e=e, exc_info=True)
             opt_info = f"{e}" if isinstance(e, ValueError) else ""
-            raise ValueError(f"error parsing ban duration string{opt_info}")
+            raise ValueError(f"error parsing ban duration string {opt_info}")
 
 
 class ModelWrapper:
@@ -1017,10 +1084,15 @@ class PlayerWrapper(ModelWrapper):
         """Player's Steam ID."""
         return self.player.steam_id
 
-    def ban(self, reason: str, duration: str = None,
+    @property
+    def notes(self) -> dict:
+        """Notes attached to player."""
+        raise NotImplementedError
+
+    def ban(self, reason: str, duration: Optional[str] = None,
             notify_players: bool = False):
-        self._adapter.ban_player(self.player, reason, duration,
-                                 notify_players)
+        self._adapter.ban_player(self.player, reason=reason, duration=duration,
+                                 notify_players=notify_players)
 
     def kick(self, reason: str, notify_players: bool = False):
         self._adapter.kick_player(self.player, reason,
@@ -1034,7 +1106,7 @@ class PlayerWrapper(ModelWrapper):
         self._adapter.revoke_player_ban(self)
 
     def revoke_session_ban(self):
-        pass
+        self._adapter.revoke_session_ban(self)
 
     def track(self):
         raise NotImplementedError
@@ -1086,6 +1158,9 @@ class SquadWrapper(ModelWrapper):
     def __repr__(self) -> str:
         return self._model.__repr__()
 
+    def reset_name(self):
+        raise NotImplementedError
+
 
 class BanWrapper(ModelWrapper):
     """Wrapper around models.Ban, providing functionality
@@ -1121,6 +1196,7 @@ class SessionBanWrapper(BanWrapper):
 
 
 class TrackingWrapper(PlayerWrapper):
+    # TODO: Should we store parsed dates?
 
     def __init__(self, player: models.Player, tracking_data: dict,
                  adapter: WebAdminAdapter):
@@ -1158,13 +1234,30 @@ class TrackingWrapper(PlayerWrapper):
     @property
     def created(self) -> datetime.datetime:
         """Creation date of tracking entry."""
-        return self.tracking["Created"]
+        return datetime.datetime.strptime(
+            self.tracking["Created"], "")
 
     @property
     def num_connects(self) -> datetime.datetime:
         """The number of times player has connected."""
-        return self.tracking["# Connects"]
+        return datetime.datetime.strptime(
+            self.tracking["# Connects"], "")
 
     @property
     def last_seen(self) -> datetime.datetime:
-        return self.tracking["Last Seen"]
+        return datetime.datetime.strptime(
+            self.tracking["Last Seen"], "")
+
+    def track(self):
+        """No effect on already tracked player."""
+        pass
+
+
+class MemberWrapper(PlayerWrapper):
+    def __init__(self, player: models.Player, member_data: dict,
+                 adapter: WebAdminAdapter):
+        super().__init__(player, adapter)
+        self._member_data = member_data
+
+    def __str__(self) -> str:
+        return f"{self._model.__str__()}, {self._member_data}"
