@@ -1,18 +1,25 @@
+"""TODO: Use dataclasses."""
+
 from __future__ import annotations
 
 import abc
 import datetime
 import sys
+from typing import Any
+from typing import Dict
+from typing import KeysView
 from typing import List
+from typing import Optional
 from typing import Tuple
 from typing import Type
 from typing import Union
 
 from logbook import Logger
 from logbook import StreamHandler
-from steam import SteamID
+from steam.steamid import SteamID
 
 from rs2wapy.adapters import adapters
+from rs2wapy.epicgamesstore import EGSID
 from rs2wapy.steam import SteamWebAPI
 
 StreamHandler(sys.stdout, level="WARNING").push_application()
@@ -43,18 +50,12 @@ HEX_COLOR_ALL_TEAM = ""
 
 
 class Team(abc.ABC):
-    HEX_COLOR = None
+    HEX_COLOR: Optional[str] = None
 
     @staticmethod
     def from_hex_color(hex_color: str) -> Type[Team]:
         return HEX_COLOR_TO_TEAM[hex_color]
 
-    @classmethod
-    def to_hex_color(cls) -> str:
-        if not cls.HEX_COLOR:
-            raise NotImplementedError("not implemented for abstract base class")
-        return TEAM_TO_HEX_COLOR[cls.HEX_COLOR]
-
     @staticmethod
     def from_team_index(index: int) -> Type[Team]:
         try:
@@ -86,48 +87,46 @@ HEX_COLOR_TO_TEAM = {
     HEX_COLOR_ALL_TEAM: AllTeam,
 }
 
-TEAM_TO_HEX_COLOR = {
-    BlueTeam: HEX_COLOR_BLUE_TEAM,
-    RedTeam: HEX_COLOR_RED_TEAM,
-    UnknownTeam: HEX_COLOR_UNKNOWN_TEAM,
-    AllTeam: HEX_COLOR_ALL_TEAM,
-}
-
-TEAM_INDEX_TO_TEAM = {
+TEAM_INDEX_TO_TEAM: Dict[int, Type[Team]] = {
     0: RedTeam,
     1: BlueTeam,
 }
 
-TEAM_TO_TEAM_INDEX = {
+TEAM_TO_TEAM_INDEX: Dict[Type[Team], int] = {
     RedTeam: 0,
     BlueTeam: 1,
 }
 
-STEAM_ID_TYPE = Union[SteamID, int, str]
-
 
+# TODO: SteamPlayer and EGSPlayer classes?
 class Player(Model):
-    def __init__(self, steam_id: STEAM_ID_TYPE, stats: dict = None,
-                 persona_name: str = None, id_intstr_base=16):
+
+    # TODO: Given int or str ID, determine Steam / EGS ID?
+    def __init__(self, ident: Union[SteamID, int, str, EGSID] = None,
+                 stats: dict = None, persona_name: str = None,
+                 id_intstr_base: int = 16):
         super().__init__()
 
+        self._steam_id = 0
+        self._egs_id = 0
+
         if not stats:
             stats = {}
         self._stats = stats
 
-        if isinstance(steam_id, SteamID):
-            self._steam_id = steam_id
-        elif isinstance(steam_id, int):
-            self._steam_id = SteamID(steam_id)
-        elif isinstance(steam_id, str):
-            self._steam_id = SteamID(int(steam_id, id_intstr_base))
+        if isinstance(ident, SteamID):
+            self._steam_id = ident
+        elif isinstance(ident, EGSID):
+            self._egs_id = ident
+        elif isinstance(ident, int):
+            self._steam_id = SteamID(ident)
+        elif isinstance(ident, str):
+            self._steam_id = SteamID(int(ident, id_intstr_base))
         else:
             raise ValueError(
-                f"invalid steam_id type: {type(steam_id)}, expected "
-                f"{STEAM_ID_TYPE}")
+                f"invalid steam_id type: {type(ident)}, expected "
+                f"{Union[SteamID, int, str]}")
 
-        if persona_name is None:
-            persona_name = SteamWebAPI().get_persona_name(self.steam_id)
         self._persona_name = persona_name
 
     @property
@@ -138,6 +137,18 @@ class Player(Model):
     def steam_id(self) -> SteamID:
         return self._steam_id
 
+    @property
+    def egs_id(self) -> EGSID:
+        return self._egs_id
+
+    @property
+    def is_steam_player(self) -> bool:
+        return self._steam_id != 0
+
+    @property
+    def is_egs_player(self) -> bool:
+        return not self.is_steam_player
+
     @property
     def name(self) -> str:
         """Player's name as stored in RS2 WebAdmin."""
@@ -145,25 +156,35 @@ class Player(Model):
             return self.stats["Player name"]
         except KeyError as ke:
             logger.debug(ke, exc_info=True)
-            logger.warn("unable to get player name")
+            logger.warn(f"unable to get player name for Steam ID {self.steam_id}")
             return ""
 
     @property
     def persona_name(self) -> str:
         """Player's Steam persona (profile) name."""
+        if self._persona_name is None and self.is_steam_player:
+            self._persona_name = SteamWebAPI().get_persona_name(self.steam_id)
         return self._persona_name
 
     def __str__(self) -> str:
-        steam_id = (self._steam_id.as_64
-                    if isinstance(self._steam_id, SteamID)
-                    else self._steam_id)
-        return f"SteamID64={steam_id}"
+        if self.is_steam_player:
+            ident = (self._steam_id.as_64
+                     if isinstance(self._steam_id, SteamID)
+                     else self._steam_id)
+            s = f"SteamID64={ident}"
+        else:
+            ident = self.egs_id.ident
+            s = f"EGSID={ident}"
+        return s
 
     def __repr__(self) -> str:
         return f"Player({self.__str__()})"
 
     def __hash__(self) -> int:
-        return self._steam_id.as_64
+        if self.is_steam_player:
+            return self._steam_id.as_64
+        else:
+            return self._egs_id.ident
 
 
 CHAT_CHANNEL_ALL_STR = "(ALL)"
@@ -221,10 +242,10 @@ class ChatMessage(Model):
             channel = f"({self._channel.to_team_str()})"
         else:
             channel = f"({self._team.__name__}) {self._channel.to_team_str()}"
-        return f"{self._timestamp.isoformat()} {self._sender} {channel}: {self._text}"
+        return f"{self.timestamp.isoformat()} {self._sender} {channel}: {self._text}"
 
     def __repr__(self) -> str:
-        return f"{__class__.__name__}({self.__str__()})"
+        return f"{type(self).__name__}({self.__str__()})"
 
     @property
     def sender(self) -> Union[Player, adapters.PlayerWrapper, str]:
@@ -251,7 +272,13 @@ class Scoreboard(abc.ABC):
         return self._stats.__str__()
 
     def __repr__(self) -> str:
-        return f"{__class__}({self.__str__()})"
+        return f"{type(self).__name__}({self.__str__()})"
+
+    def keys(self) -> KeysView:
+        return self._stats.keys()
+
+    def __getitem__(self, item: Any):
+        return self._stats.__getitem__(item)
 
 
 class PlayerScoreboard(Scoreboard):
@@ -272,6 +299,19 @@ class CurrentGame(Model):
         self._info = info
         self._rules = rules
 
+    def __str__(self) -> str:
+        # TODO: implement keys() and __getitem__?
+        all_info = {}
+        all_info.update(self._info)
+        all_info.update(self._rules)
+        all_info.update(self._player_scoreboard)
+        all_info.update(self._team_scoreboard)
+        return str(all_info)
+
+    def __repr__(self) -> str:
+        # TODO:
+        return self.__str__()
+
     @property
     def player_scoreboard(self) -> PlayerScoreboard:
         return self._player_scoreboard
@@ -346,7 +386,7 @@ class MapCycle(Model):
         return f"{self._maps}"
 
     def __repr__(self) -> str:
-        return f"{__class__.__name__}({self.__str__()})"
+        return f"{type(self).__name__}({self.__str__()})"
 
 
 class Squad(Model):
@@ -360,7 +400,7 @@ class Squad(Model):
         return f"team={self._team}, number={self._number}, name={self._name}"
 
     def __repr__(self) -> str:
-        return f"{__class__.__name__}({self.__str__()})"
+        return f"{type(self).__name__}({self.__str__()})"
 
     def __hash__(self) -> int:
         return self._number
@@ -386,13 +426,22 @@ class Ban(Model):
         return self._reason
 
     @property
-    def until(self) -> datetime.datetime:
+    def until(self) -> Optional[datetime.datetime]:
         """Ban expiration date. If None, the ban is permanent."""
         return self._until
 
+    @property
+    def expired(self) -> bool:
+        """True if ban has expired."""
+        if self.until is not None:
+            return self.until > datetime.datetime.now()
+        else:
+            return False
+
     @staticmethod
     def _parse_until(until: str) -> datetime.datetime:
         """Ban expiration date str to datetime.datetime object."""
+        # TODO:
         return datetime.datetime.strptime(until, "")
 
 
