@@ -1,5 +1,7 @@
 from .parsing import RS2WebAdminResponseParser
+from .parsing import logger
 
 __all__ = [
     "RS2WebAdminResponseParser",
+    "logger",
 ]
