@@ -1,22 +1,26 @@
 """
 Provides utilities for parsing responses from
 Rising Storm 2: Vietnam WebAdmin.
+
+TODO: This module is growing fast. Needs refactoring.
 """
 from __future__ import annotations
 
 import re
 import sys
-from typing import List
+from typing import List, Dict
 from typing import Sequence
 from typing import Tuple
+from typing import Union
 
-import steam
 from bs4 import BeautifulSoup
 from logbook import Logger
 from logbook import StreamHandler
+from steam.steamid import SteamID
 
 import rs2wapy.models as models
 from rs2wapy.adapters import adapters
+from rs2wapy.epicgamesstore import EGSID
 from rs2wapy.steam import SteamWebAPI
 
 StreamHandler(sys.stdout, level="WARNING").push_application()
@@ -31,6 +35,7 @@ TEAM_INDEX_KEY = "\xa0"
 
 
 class RS2WebAdminResponseParser:
+    # TODO: Refactor magic numbers.
 
     def __init__(self, encoding: str = None):
         if not encoding:
@@ -53,38 +58,6 @@ class RS2WebAdminResponseParser:
             cm.append(self.parse_chat_message(div))
         return cm
 
-    @staticmethod
-    def parse_chat_message(div: BeautifulSoup) -> models.ChatMessage:
-        teamcolor = str(div.find(
-            "span", attrs={"class": "teamcolor"}).get("style"))
-        if not teamcolor:
-            logger.error(
-                "no teamcolor in chat message div={div}", div=div)
-        else:
-            try:
-                teamcolor = re.match(TEAMCOLOR_PATTERN, teamcolor).groups()[0]
-            except IndexError as ie:
-                logger.error("error getting teamcolor: {e}", e=ie)
-
-        teamnotice = div.find("span", attrs={"class": "teamnotice"})
-        if teamnotice:
-            teamnotice = teamnotice.text
-
-        name = div.find("span", attrs={"class": "username"})
-        if name:
-            name = name.text
-
-        msg = div.find("span", attrs={"class": "message"})
-        if msg:
-            msg = msg.text
-
-        return models.ChatMessage(
-            sender=name,
-            text=msg,
-            team=models.Team.from_hex_color(str(teamcolor)),
-            channel=models.ChatChannel.from_teamnotice(teamnotice)
-        )
-
     def parse_access_policy(self, resp: bytes,
                             encoding: str = None) -> List[str]:
         parsed_html = self.parse_html(resp, encoding)
@@ -284,24 +257,24 @@ class RS2WebAdminResponseParser:
         # We use 'Unique ID' column here instead of 'Steam ID'
         # column because 'Steam ID' column is sometimes not filled.
         # 'Unique ID' is just a hex-string of the user's SteamID64
-        # anyway.
+        # anyway. EGS players are also handled in similar way.
         id_index = player_headers.index(UNIQUE_ID_KEY)
 
         players = []
-        id_to_stats = {}
+        id_to_stats: Dict[Union[SteamID, EGSID], Dict] = {}
         steam_ids = []
+        egs_ids = []
 
         for player_row in player_table:
-            steam_id = player_row[id_index]
+            unique_id = player_row[id_index]
 
             try:
-                steam_id = int(steam_id, 16)
+                unique_id = int(unique_id, 16)
             except ValueError as ve:
-                logger.error("unable to convert Unique ID to SteamID64")
+                logger.error("unable to convert Unique ID '{uid}' to int, skipping",
+                             uid=unique_id)
                 logger.exception(ve)
-                steam_id = 0
-
-            steam_ids.append(steam.SteamID(steam_id))
+                continue
 
             stats = {
                 key: value for key, value in zip(
@@ -309,7 +282,16 @@ class RS2WebAdminResponseParser:
                 if key.lower() != "actions"
             }
 
-            id_to_stats[steam_id] = stats
+            # Very likely a Steam ID.
+            if len(str(unique_id)) == 17:
+                steam_id = SteamID(unique_id)
+                steam_ids.append(steam_id)
+                id_to_stats[steam_id] = stats
+            # Very likely an EGS ID.
+            else:
+                egs_id = EGSID(unique_id)
+                egs_ids.append(egs_id)
+                id_to_stats[egs_id] = stats
 
         persona_names = SteamWebAPI().get_persona_names(
             steam_ids=steam_ids
@@ -327,11 +309,17 @@ class RS2WebAdminResponseParser:
 
             p_stats = id_to_stats[steam_id]
 
-            player = models.Player(
-                steam_id=steam_id,
-                stats=p_stats,
-                persona_name=persona_name,
-            )
+            player = models.Player(ident=steam_id, stats=p_stats, persona_name=persona_name)
+
+            players.append(adapters.PlayerWrapper(
+                player=player,
+                adapter=adapter,
+            ))
+
+        for egs_id in egs_ids:
+            p_stats = id_to_stats[egs_id]
+
+            player = models.Player(ident=egs_id, stats=p_stats)
 
             players.append(adapters.PlayerWrapper(
                 player=player,
@@ -343,12 +331,10 @@ class RS2WebAdminResponseParser:
     def parse_hash_alg(self, resp: bytes) -> str:
         parsed_html = self.parse_html(resp)
         login_script = parsed_html.find("script", text=re.compile(r".*hashAlg.*"))
+        match = re.search(r'.*var\shashAlg\s=\s"(.*)";.*', str(login_script))
         alg = ""
-        try:
-            alg = re.search(r'.*var\shashAlg\s=\s"(.*)";.*', str(login_script))
-            alg = alg.group(1)
-        except AttributeError as ae:
-            logger.error("unable to parse hash alg: {e}", e=ae)
+        if match:
+            alg = match.group(1)
         return alg
 
     def parse_map_list_indices(self, resp) -> dict:
@@ -407,11 +393,11 @@ class RS2WebAdminResponseParser:
 
         squads = []
         for row in squads_table:
-            team = models.UnknownTeam
             try:
                 team_idx = int(row[0].strip())
                 team = models.Team.from_team_index(team_idx)
             except Exception as e:
+                team = models.UnknownTeam
                 logger.error("unable to parse squad team: {e}", e=e)
 
             number = -1
@@ -452,13 +438,28 @@ class RS2WebAdminResponseParser:
         tb_row_entries = tracking_tbody.find_all("tr")
         tracking_rows = self._parse_table(tb_row_entries)
 
+        # Unique ID field is more reliable as Steam ID
+        # might be missing in the table in some cases.
+        id_index = tracking_headers.index("Unique ID")
+
+        steam_ids = [SteamID(int(row[id_index], 16))
+                     for row in tracking_rows]
+        persona_names = SteamWebAPI().get_persona_names(
+            steam_ids=steam_ids
+        )
+
         tracking_wrappers = []
         for row in tracking_rows:
-            steam_id = row[3]
-            player = models.Player(
-                steam_id=steam_id,
-                id_intstr_base=10,
-            )
+            steam_id = SteamID(int(row[id_index], 16))
+
+            persona_name = ""
+            try:
+                persona_name = persona_names[steam_id]
+            except KeyError as ke:
+                logger.warn("cannot get persona name for Steam ID: {sid}: {e}",
+                            sid=steam_id, e=ke)
+
+            player = models.Player(ident=steam_id, persona_name=persona_name)
 
             tracking_data = {
                 key: value for key, value in zip(
@@ -474,12 +475,147 @@ class RS2WebAdminResponseParser:
 
         return tracking_wrappers
 
+    def parse_fvri(self, resp: bytes):
+        """Parse first visible row index from response."""
+        parsed_html = self.parse_html(resp)
+        return parsed_html.find(
+            "input", attrs={"id": "__FirstVisibleRowIndex"}).get("value")
+
+    def parse_has_next_page(self, resp: bytes):
+        """Return True if next page button is enabled."""
+        parsed_html = self.parse_html(resp)
+        np_button = parsed_html.find(
+            "button", attrs={"id": "__NextPage"})
+        if not np_button:
+            return False
+        button_disabled = np_button.has_attr("disabled")
+        return not button_disabled
+
+    def parse_player_id_key(
+            self, resp: bytes,
+            player: Union[models.Player, adapters.PlayerWrapper]
+    ) -> Tuple[int, str]:
+        """Parse "hidden" player key and ID from players table."""
+        parsed_html = self.parse_html(resp)
+        trs = parsed_html.find_all("tr")
+        player_id = -1
+        player_key = ""
+
+        for tr in trs:
+            div = tr.find("div")
+            if not div:
+                continue
+
+            # Find x from <input id = "__PlayerId_x">.
+            x = div.find("input").get("id").split("_")[-1]
+            div_player_key = div.find(
+                "input", attrs={"id": f"__PlayerKey_{x}"}).get("value")
+            div_player_name = div.find(
+                "input", attrs={"id": f"__PlayerName_{x}"}).get("value")
+
+            div_player_steam_id = div_player_key.split("_")[1].strip().lower()
+            steam_id = player.steam_id.as_64
+            # Match leading zeros of strings to compare.
+            steam_id = f"{steam_id:#0{len(div_player_steam_id)}x}"
+            steam_id = steam_id.strip().lower()
+
+            if (steam_id == div_player_steam_id
+                    and (div_player_name == player.name)):
+                player_id = int(div.find(
+                    "input", attrs={"id": f"__PlayerId_{x}"}).get("value"))
+                player_key = div_player_key
+                break
+
+        return player_id, player_key
+
+    def parse_members(self, resp: bytes,
+                      adapter: adapters.WebAdminAdapter
+                      ) -> List[adapters.MemberWrapper]:
+        parsed_html = self.parse_html(resp)
+        members_table = parsed_html.find("table", attrs={"id": "members"})
+        members_headers = [th.text for th in members_table.find_all("th")]
+        tracking_tbody = members_table.find("tbody")
+        tb_row_entries = tracking_tbody.find_all("tr")
+        members_rows = self._parse_table(tb_row_entries)
+
+        id_index = members_headers.index("Unique ID")
+
+        steam_ids = [SteamID(int(row[id_index], 16))
+                     for row in members_rows]
+
+        persona_names = SteamWebAPI().get_persona_names(
+            steam_ids=steam_ids
+        )
+
+        member_wrappers = []
+        for row in members_rows:
+            steam_id = SteamID(int(row[id_index], 16))
+
+            persona_name = ""
+            try:
+                if not steam_id.is_valid():
+                    raise KeyError("invalid SteamID")
+                persona_name = persona_names[steam_id]
+            except KeyError as ke:
+                logger.warn("cannot get persona name for Steam ID: {sid}: {e}",
+                            sid=steam_id, e=ke)
+
+            player = models.Player(ident=steam_id, persona_name=persona_name)
+
+            member_data = {
+                key: value for key, value in zip(
+                    members_headers, row)
+                if key.lower() != "actions"
+            }
+
+            member_wrappers.append(adapters.MemberWrapper(
+                player=player,
+                member_data=member_data,
+                adapter=adapter,
+            ))
+
+        return member_wrappers
+
+    @staticmethod
+    def parse_chat_message(div: BeautifulSoup) -> models.ChatMessage:
+        teamcolor = str(div.find(
+            "span", attrs={"class": "teamcolor"}).get("style"))
+        if not teamcolor:
+            logger.error(
+                "no teamcolor in chat message div={div}", div=div)
+        else:
+            try:
+                match = re.match(TEAMCOLOR_PATTERN, teamcolor)
+                if match:
+                    teamcolor = match.groups()[0]
+            except IndexError as ie:
+                logger.error("error getting teamcolor: {e}", e=ie)
+
+        teamnotice = div.find("span", attrs={"class": "teamnotice"})
+        if teamnotice:
+            teamnotice = teamnotice.text
+
+        name = div.find("span", attrs={"class": "username"})
+        if name:
+            name = name.text
+
+        msg = div.find("span", attrs={"class": "message"})
+        if msg:
+            msg = msg.text
+
+        return models.ChatMessage(
+            sender=name,
+            text=msg,
+            team=models.Team.from_hex_color(str(teamcolor)),
+            channel=models.ChatChannel.from_teamnotice(teamnotice)
+        )
+
     @staticmethod
     def _parse_table(row_elements: Sequence) -> List[List[str]]:
         all_cols = []
         for row in row_elements:
             cols = row.find_all("td")
-            cols = [ele.text.strip() for ele in cols]
+            cols = [ele.get_text(" ").strip() for ele in cols]
             if not cols:
                 continue
             all_cols.append(cols)
