@@ -1,3 +1,5 @@
+"""High level API."""
+
 from __future__ import annotations
 
 import sys
@@ -12,6 +14,7 @@ from logbook import StreamHandler
 from rs2wapy.adapters import PlayerWrapper
 from rs2wapy.adapters import WebAdminAdapter
 from rs2wapy.adapters.adapters import BanWrapper
+from rs2wapy.adapters.adapters import MemberWrapper
 from rs2wapy.adapters.adapters import SquadWrapper
 from rs2wapy.adapters.adapters import TrackingWrapper
 from rs2wapy.models import AccessPolicy
@@ -132,7 +135,14 @@ class RS2WebAdmin:
         # return self._adapter.get_banned_players()
 
     def get_tracked_players(self) -> List[TrackingWrapper]:
-        """Return tracked players."""
+        """Return tracked players.
+
+        WARNING: This method is extremely slow for servers
+        with large tracking databases.
+
+        TODO: Can we leverage asyncio?
+        TODO: Background collection like chat?
+        """
         return self._adapter.get_tracked_players()
 
     def get_access_policies(self) -> List[AccessPolicy]:
@@ -175,8 +185,8 @@ class RS2WebAdmin:
         :param notify_players:
             If True, notify players on the server.
         """
-        self._adapter.ban_player(player, reason, duration,
-                                 notify_players)
+        self._adapter.ban_player(player, reason=reason, duration=duration,
+                                 notify_players=notify_players)
 
     def kick_player(self, player: Union[Player, PlayerWrapper],
                     reason: str, notify_players: bool = False):
@@ -224,3 +234,6 @@ class RS2WebAdmin:
 
     def set_advertisement_interval(self, ad_interval: int):
         raise NotImplementedError
+
+    def get_members(self) -> List[MemberWrapper]:
+        return self._adapter.get_members()
